@@ -1,15 +1,17 @@
+const char rcsid_scc_c[] = "@(#)$KmKId: scc.c,v 1.52 2021-08-17 00:08:36+00 kentd Exp $";
+
 /************************************************************************/
 /*			KEGS: Apple //gs Emulator			*/
-/*			Copyright 2002 by Kent Dickey			*/
+/*			Copyright 2002-2021 by Kent Dickey		*/
 /*									*/
-/*		This code is covered by the GNU GPL			*/
+/*	This code is covered by the GNU GPL v3				*/
+/*	See the file COPYING.txt or https://www.gnu.org/licenses/	*/
+/*	This program is provided with no warranty			*/
 /*									*/
 /*	The KEGS web page is kegs.sourceforge.net			*/
 /*	You may contact the author at: kadickey@alumni.princeton.edu	*/
 /************************************************************************/
 
-const char rcsid_scc_c[] = "@(#)$KmKId: scc.c,v 1.44 2004-12-03 17:33:40-05 kentd Exp $";
-
 #include "defc.h"
 
 extern int Verbose;
@@ -19,8 +21,8 @@ extern int g_raw_serial;
 extern int g_serial_out_masking;
 extern int g_irq_pending;
 
-/* my scc port 0 == channel A = slot 1 = c039/c03b */
-/*        port 1 == channel B = slot 2 = c038/c03a */
+/* scc	port 0 == channel A = slot 1 = c039/c03b */
+/*	port 1 == channel B = slot 2 = c038/c03a */
 
 #include "scc.h"
 #define SCC_R14_DPLL_SOURCE_BRG		0x100
@@ -72,6 +74,7 @@ scc_init()
 		scc_ptr->modem_plus_mode = 0;
 		scc_ptr->modem_s0_val = 0;
 		scc_ptr->modem_cmd_len = 0;
+		scc_ptr->modem_out_portnum = 23;
 		scc_ptr->modem_cmd_str[0] = 0;
 		for(j = 0; j < 2; j++) {
 			scc_ptr->telnet_local_mode[j] = 0;
@@ -175,19 +178,10 @@ void
 scc_regen_clocks(int port)
 {
 	Scc	*scc_ptr;
-	double	br_dcycs, tx_dcycs, rx_dcycs;
-	double	rx_char_size, tx_char_size;
+	double	br_dcycs, tx_dcycs, rx_dcycs, rx_char_size, tx_char_size;
 	double	clock_mult;
-	word32	reg4;
-	word32	reg14;
-	word32	reg11;
-	word32	br_const;
-	word32	baud;
-	word32	max_diff;
-	word32	diff;
-	int	state;
-	int	baud_entries;
-	int	pos;
+	word32	reg4, reg11, reg14, br_const, max_diff, diff;
+	int	baud, state, baud_entries, pos;
 	int	i;
 
 	/*	Always do baud rate generator */
@@ -301,7 +295,7 @@ scc_regen_clocks(int port)
 		scc_serial_win_change_params(port);
 #endif
 	} else {
-		scc_socket_change_params(port);
+		// scc_socket_change_params(port);
 	}
 }
 
@@ -429,7 +423,7 @@ do_scc_event(int type, double dcycs)
 	} else if(type == SCC_TX_EVENT) {
 		scc_ptr->tx_event_pending = 0;
 		scc_ptr->tx_buf_empty = 1;
-		scc_handle_tx_event(port, dcycs);
+		scc_handle_tx_event(port);
 	} else if(type == SCC_RX_EVENT) {
 		scc_ptr->rx_event_pending = 0;
 		scc_maybe_rx_event(port, dcycs);
@@ -597,7 +591,6 @@ scc_read_reg(int port, double dcycs)
 			}
 			if(wr9 & 0x10) {
 				/* wr9 status high */
-				
 			}
 #endif
 		}
@@ -831,7 +824,6 @@ scc_write_reg(int port, word32 val, double dcycs)
 		case 0x2:
 		case 0x3:
 			break;
-		
 		case 0x4:	/* DPLL source is BR gen */
 			val |= SCC_R14_DPLL_SOURCE_BRG;
 			break;
@@ -942,7 +934,6 @@ scc_evaluate_ints(int port)
 	}
 }
 
-
 void
 scc_maybe_rx_event(int port, double dcycs)
 {
@@ -974,14 +965,14 @@ scc_maybe_rx_event(int port, double dcycs)
 	scc_ptr->rx_queue[depth] = scc_ptr->in_buf[in_rdptr];
 	scc_ptr->in_rdptr = (in_rdptr + 1) & (SCC_INBUF_SIZE - 1);
 	scc_ptr->rx_queue_depth = depth + 1;
-	scc_maybe_rx_int(port, dcycs);
+	scc_maybe_rx_int(port);
 	rx_dcycs = scc_ptr->rx_dcycs;
 	scc_ptr->rx_event_pending = 1;
 	add_event_scc(dcycs + rx_dcycs, SCC_MAKE_EVENT(port, SCC_RX_EVENT));
 }
 
 void
-scc_maybe_rx_int(int port, double dcycs)
+scc_maybe_rx_int(int port)
 {
 	Scc	*scc_ptr;
 	int	depth;
@@ -1010,7 +1001,7 @@ scc_clr_rx_int(int port)
 }
 
 void
-scc_handle_tx_event(int port, double dcycs)
+scc_handle_tx_event(int port)
 {
 	Scc	*scc_ptr;
 	int	tx_int_mode;
@@ -1109,14 +1100,14 @@ scc_add_to_readbufv(int port, double dcycs, const char *fmt, ...)
 {
 	va_list	ap;
 	char	*bufptr;
-	int	ret, len, c;
+	int	len, c;
 	int	i;
 
 	va_start(ap, fmt);
 	bufptr = malloc(4096);
 	bufptr[0] = 0;
-	ret = vsnprintf(bufptr, 4090, fmt, ap);
-	len = strlen(bufptr);
+	vsnprintf(bufptr, 4090, fmt, ap);
+	len = (int)strlen(bufptr);
 	for(i = 0; i < len; i++) {
 		c = bufptr[i];
 		if(c == 0x0a) {
@@ -1128,7 +1119,7 @@ scc_add_to_readbufv(int port, double dcycs, const char *fmt, ...)
 }
 
 void
-scc_transmit(int port, word32 val, double dcycs)
+scc_transmit(int port, word32 val)
 {
 	Scc	*scc_ptr;
 	int	out_wrptr;
@@ -1164,11 +1155,11 @@ scc_transmit(int port, word32 val, double dcycs)
 		val = val & 0x7f;
 	}
 
-	scc_add_to_writebuf(port, val, dcycs);
+	scc_add_to_writebuf(port, val);
 }
 
 void
-scc_add_to_writebuf(int port, word32 val, double dcycs)
+scc_add_to_writebuf(int port, word32 val)
 {
 	Scc	*scc_ptr;
 	int	out_wrptr;
@@ -1227,7 +1218,7 @@ scc_read_data(int port, double dcycs)
 		}
 		scc_ptr->rx_queue_depth = depth - 1;
 		scc_maybe_rx_event(port, dcycs);
-		scc_maybe_rx_int(port, dcycs);
+		scc_maybe_rx_int(port);
 	}
 
 	scc_printf("SCC read %04x: ret %02x, depth:%d\n", 0xc03b-port, ret,
@@ -1252,7 +1243,7 @@ scc_write_data(int port, word32 val, double dcycs)
 		/* local loopback! */
 		scc_add_to_readbuf(port, val, dcycs);
 	} else {
-		scc_transmit(port, val, dcycs);
+		scc_transmit(port, val);
 	}
 	scc_try_to_empty_writebuf(port, dcycs);
 
